# pylint: disable=fixme,too-many-locals,too-many-branches

from __future__ import annotations

__all__ = (
    "Charclass",
    "DIGIT",
    "DOT",
    "NONDIGITCHAR",
    "NONSPACECHAR",
    "NONWORDCHAR",
    "NULLCHARCLASS",
    "SPACECHAR",
    "WORDCHAR",
    "escapes",
    "negate",
    "shorthand",
    "repartition",
)

from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, List, Mapping, Tuple

NUM_UNICODE_CHARS = (1 << 16) + (1 << 20)


def negate(ord_ranges: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    u = 0
    negated = []
    for ord_range in ord_ranges:
        if u < ord_range[0]:
            negated.append((u, ord_range[0] - 1))
        u = ord_range[1] + 1
    if u < NUM_UNICODE_CHARS - 1:
        negated.append((u, NUM_UNICODE_CHARS - 1))
    return negated


def collapse_ord_ranges(ord_ranges: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """
    Assume all existing ord ranges are sorted, and also disjoint
    So no cases of [[12, 17], [2, 3]] or [[4, 6], [7, 8]].
    """
    collapsed: List[Tuple[int, int]] = []

    for ord_range in sorted(ord_ranges):
        if not collapsed or collapsed[-1][1] + 1 < ord_range[0]:
            collapsed.append(ord_range)
        elif ord_range[1] > collapsed[-1][1]:
            # merge into previous
            collapsed[-1] = (collapsed[-1][0], ord_range[1])

    return collapsed


@dataclass(frozen=True, init=False)
class Charclass:
    """
    A `Charclass` is basically a `frozenset` of symbols.
    A `Charclass` with the `negated` flag set is assumed
    to contain every symbol that is in the alphabet of all symbols but not
    explicitly listed inside the frozenset. e.g. [^a]. This is very handy
    if the full alphabet is extremely large, but also requires dedicated
    combination functions.
    """

    ord_ranges: List[Tuple[int, int]]
    negated: bool

    def __init__(
        self, ranges: str | Tuple[Tuple[str, str], ...] = "", negated: bool = False
    ):
        if isinstance(ranges, str):
            ranges = tuple((char, char) for char in ranges)
        if not isinstance(ranges, tuple):
            raise TypeError(f"Bad ranges: {ranges!r}")
        for r in ranges:
            if len(r) != 2 or r[0] > r[1]:
                raise ValueError(f"Bad range: {r!r}")
            for char in r:
                if not isinstance(char, str):
                    raise TypeError(f"Can't put {char!r} in a `Charclass`", char)
                if len(char) != 1:
                    raise ValueError("`Charclass` can only contain single chars", char)

        # Rebalance ranges!
        ord_ranges = [(ord(first), ord(last)) for first, last in ranges]
        ord_ranges = collapse_ord_ranges(ord_ranges)

        object.__setattr__(self, "ord_ranges", tuple(ord_ranges))
        object.__setattr__(self, "negated", negated)

    def __lt__(self, other: Charclass, /) -> bool:
        if self.negated < other.negated:
            return True
        if (
            self.negated == other.negated
            and self.ord_ranges[0][0] < other.ord_ranges[0][0]
        ):
            return True
        return False

    def __eq__(self, other: object, /) -> bool:
        return (
            isinstance(other, Charclass)
            and self.ord_ranges == other.ord_ranges
            and self.negated == other.negated
        )

    def __hash__(self, /) -> int:
        return hash((self.ord_ranges, self.negated))

    # These are the characters carrying special meanings when they appear
    # "outdoors" within a regular expression. To be interpreted literally, they
    # must be escaped with a backslash.
    allSpecial: ClassVar[frozenset[str]] = frozenset("\\[]|().?*+{}")

    # These are the characters carrying special meanings when they appear
    # INSIDE a character class (delimited by square brackets) within a regular
    # expression. To be interpreted literally, they must be escaped with a
    # backslash. Notice how much smaller this class is than the one above; note
    # also that the hyphen and caret do NOT appear above.
    classSpecial: ClassVar[frozenset[str]] = frozenset("\\[]^-")

    def __str__(self, /) -> str:
        # pylint: disable=too-many-return-statements

        # e.g. \w
        if self in shorthand:
            return shorthand[self]

        # e.g. [^a]
        if self.negated:
            return f"[^{self.escape()}]"

        # single character, not contained inside square brackets.
        if len(self.ord_ranges) == 1 and self.ord_ranges[0][0] == self.ord_ranges[0][1]:
            u = self.ord_ranges[0][0]
            char = chr(u)

            # e.g. if char is "\t", return "\\t"
            if char in escapes:
                return escapes[char]

            if char in Charclass.allSpecial:
                return f"\\{char}"

            # If char is an ASCII control character, don't print it directly,
            # return a hex escape sequence e.g. "\\x00". Note that this
            # includes tab and other characters already handled above
            if 0 <= u <= 0x1F or u == 0x7F:
                return f"\\x{u:02x}"

            return char

        # multiple characters (or possibly 0 characters)
        return f"[{self.escape()}]"

    def escape(self, /) -> str:
        def escape_char(char: str, /) -> str:
            if char in Charclass.classSpecial:
                return f"\\{char}"
            if char in escapes:
                return escapes[char]

            # If char is an ASCII control character, don't print it directly,
            # return a hex escape sequence e.g. "\\x00". Note that this
            # includes tab and other characters already handled above
            if 0 <= ord(char) <= 0x1F or ord(char) == 0x7F:
                return f"\\x{ord(char):02x}"

            return char

        output = ""

        for first_u, last_u in self.ord_ranges:
            # there's no point in putting a range when the whole thing is
            # 3 characters or fewer. "abc" -> "abc" but "abcd" -> "a-d"
            if last_u <= first_u + 2:
                # "a" or "ab" or "abc" or "abcd"
                for u in range(first_u, last_u + 1):
                    output += escape_char(chr(u))
            else:
                # "a-b" or "a-c" or "a-d"
                output += escape_char(chr(first_u)) + "-" + escape_char(chr(last_u))

        return output

    def __repr__(self, /) -> str:
        sign = "~" if self.negated else ""
        ranges = tuple(
            (chr(first_u), chr(last_u)) for (first_u, last_u) in self.ord_ranges
        )
        return f"{sign}Charclass({ranges!r})"

    def reduce(self, /) -> Charclass:
        # `Charclass`es cannot be reduced.
        return self

    def empty(self, /) -> bool:
        return not self.ord_ranges and not self.negated

    # set operations
    def negate(self, /) -> Charclass:
        """
        Negate the current `Charclass`. e.g. [ab] becomes [^ab]. Call
        using "charclass2 = ~charclass1"
        """
        ranges = tuple(
            (chr(first_u), chr(last_u)) for (first_u, last_u) in self.ord_ranges
        )
        return Charclass(ranges, negated=not self.negated)

    def __invert__(self, /) -> Charclass:
        return self.negate()

    def get_chars(self, /) -> Iterator[str]:
        """
        Use this with caution, it can iterate over 1,000,000+ characters
        """
        for first_u, last_u in self.ord_ranges:
            for u in range(first_u, last_u + 1):
                yield chr(u)

    def num_chars(self, /) -> int:
        num = 0
        for first_u, last_u in self.ord_ranges:
            num += last_u + 1 - first_u
        return NUM_UNICODE_CHARS - num if self.negated else num

    def accepts(self, char: str, /) -> bool:
        u = ord(char)
        for first_u, last_u in self.ord_ranges:
            if first_u <= u <= last_u:
                return not self.negated
        return self.negated

    def reversed(self, /) -> Charclass:
        return self

    def union(self, other: Charclass, /) -> Charclass:
        # TODO: make this able to efficiently unite many Charclasses at once,
        # again
        self_ord_ranges = list(self.ord_ranges)
        if self.negated:
            self_ord_ranges = negate(self_ord_ranges)

        other_ord_ranges = list(other.ord_ranges)
        if other.negated:
            other_ord_ranges = negate(other_ord_ranges)

        new_ord_ranges = []
        new_ord_ranges.extend(self_ord_ranges)
        new_ord_ranges.extend(other_ord_ranges)
        new_ord_ranges = collapse_ord_ranges(new_ord_ranges)

        new_negated = self.negated or other.negated
        if new_negated:
            new_ord_ranges = negate(new_ord_ranges)
        new_ranges = tuple(
            (chr(first_u), chr(last_u)) for (first_u, last_u) in new_ord_ranges
        )
        return Charclass(new_ranges, new_negated)

    __or__ = union

    def issubset(self, other: Charclass, /) -> bool:
        return self | other == other

    def intersection(self, other: Charclass, /) -> Charclass:
        # TODO: is this actually efficient?
        # TODO: make this able to efficiently intersect many Charclasses at once,
        # again
        return ~(~self | ~other)

    __and__ = intersection


# Standard character classes
WORDCHAR = Charclass("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz")
DIGIT = Charclass("0123456789")
SPACECHAR = Charclass("\t\n\v\f\r ")

# This `Charclass` expresses "no possibilities at all"
# and can never match anything.
NULLCHARCLASS = Charclass()

NONWORDCHAR = ~WORDCHAR
NONDIGITCHAR = ~DIGIT
NONSPACECHAR = ~SPACECHAR
DOT = ~NULLCHARCLASS

# Textual representations of standard character classes
shorthand: Mapping[Charclass, str] = {
    WORDCHAR: "\\w",
    DIGIT: "\\d",
    SPACECHAR: "\\s",
    NONWORDCHAR: "\\W",
    NONDIGITCHAR: "\\D",
    NONSPACECHAR: "\\S",
    DOT: ".",
}

# Characters which users may escape in a regex instead of inserting them
# literally. In ASCII order:
escapes: Mapping[str, str] = {
    "\t": "\\t",  # tab
    "\n": "\\n",  # line feed
    "\v": "\\v",  # vertical tab
    "\f": "\\f",  # form feed
    "\r": "\\r",  # carriage return
}


def repartition(
    charclasses: Iterable[Charclass],
) -> Mapping[Charclass, Iterable[Charclass]]:
    """
    Accept an iterable of `Charclass`es which may overlap somewhat.
    Construct a minimal collection of `Charclass`es which partition the space
    of all possible characters and can be combined to create all of the
    originals.
    Return a map from each original `Charclass` to its constituent pieces.
    """
    ord_range_boundaries = set()
    for charclass in charclasses:
        for first_u, last_u in charclass.ord_ranges:
            ord_range_boundaries.add(first_u)
            ord_range_boundaries.add(last_u + 1)
    ord_range_boundaries_2 = sorted(ord_range_boundaries)

    ord_ranges = []
    for i, ord_range_boundary in enumerate(ord_range_boundaries_2):
        if i + 1 < len(ord_range_boundaries_2):
            ord_ranges.append((ord_range_boundary, ord_range_boundaries_2[i + 1] - 1))

    # Group all of the possible ranges by "signature".
    # A signature is a tuple of Booleans telling us which character classes
    # a particular range is mentioned in.
    # (Whether it's *accepted* is actually not relevant.)
    signatures: Dict[Tuple[bool, ...], List[Tuple[int, int]]] = {}
    for ord_range in ord_ranges:
        signature = []
        for charclass in charclasses:
            ord_range_in_charclass = False
            for x in charclass.ord_ranges:
                if x[0] <= ord_range[0] and ord_range[1] <= x[1]:
                    ord_range_in_charclass = True
                    break
            signature.append(ord_range_in_charclass)
        signature2 = tuple(signature)
        if signature2 not in signatures:
            signatures[signature2] = []
        signatures[signature2].append(ord_range)

    # From the signatures we can gather the new Charclasses
    newcharclasses = []
    newcharclasses.append(
        ~Charclass(
            tuple((chr(first_u), chr(last_u)) for (first_u, last_u) in ord_ranges)
        )
    )
    for ord_ranges2 in signatures.values():
        newcharclasses.append(
            Charclass(
                tuple((chr(first_u), chr(last_u)) for (first_u, last_u) in ord_ranges2)
            )
        )

    # Now compute the breakdowns
    partition: Dict[Charclass, List[Charclass]] = {}
    for charclass in charclasses:
        partition[charclass] = []
        for newcharclass in newcharclasses:
            if newcharclass.issubset(charclass):
                partition[charclass].append(newcharclass)

    return partition



from __future__ import annotations

__all__ = (
    "Bound",
    "INF",
    "Multiplier",
    "PLUS",
    "Pattern",
    "QM",
    "STAR",
    "parse",
    "Fsm",
    "EPSILON",
    "NULL",
    "Charclass",
)

from .bound import INF, Bound
from .fsm import EPSILON, NULL, Charclass, Fsm
from .multiplier import PLUS, QM, STAR, Multiplier
from .parse import parse
from .rxelems import Pattern



from __future__ import annotations

__all__ = (
    "parse",
    "NoMatch",
)

from typing import Collection, Tuple, TypeVar

from .bound import INF, Bound
from .charclass import (
    DIGIT,
    NONDIGITCHAR,
    NONSPACECHAR,
    NONWORDCHAR,
    SPACECHAR,
    WORDCHAR,
    Charclass,
    escapes,
    shorthand,
)
from .multiplier import ONE, Multiplier, symbolic
from .rxelems import Conc, Mult, Pattern

# Currently many statements are grouped by `try/except NoMatch` in order to try
# multiple matching functions in sequence. They can be refactored into smaller
# functions to remove this suppression.
# pylint: disable=too-many-try-statements

T_co = TypeVar("T_co", covariant=True)


class NoMatch(Exception):
    """
    Thrown when parsing fails.
    Almost always caught and almost never fatal
    """


MatchResult = Tuple[T_co, int]


def read_until(string: str, i: int, stop_char: str) -> MatchResult[str]:
    start = i
    while True:
        if i >= len(string):
            raise NoMatch
        if string[i] == stop_char:
            break
        i += 1
    return string[start:i], i + 1


def static(haystack: str, i: int, needle: str) -> int:
    j = i + len(needle)
    if haystack[i:j] == needle:
        return j
    raise NoMatch


def select_static(haystack: str, i: int, *needles: str) -> MatchResult[str]:
    for needle in needles:
        j = i + len(needle)
        if haystack[i:j] == needle:
            return needle, j
    raise NoMatch


def unescape_hex(string: str, i: int) -> MatchResult[str]:
    """Turn e.g. "\\x40" into "@". Exactly two hex digits"""
    hex_digits = "0123456789AaBbCcDdEeFf"

    j = static(string, i, "\\x")

    hex1 = string[j]  # e.g. "4"
    if hex1 not in hex_digits:
        raise NoMatch
    j += len(hex1)

    hex2 = string[j]  # e.g. "0"
    if hex2 not in hex_digits:
        raise NoMatch
    j += len(hex2)

    codepoint = int(hex1 + hex2, 16)  # e.g. 64
    char = chr(codepoint)  # "@"
    return char, j


def match_internal_char(string: str, i: int) -> MatchResult[str]:
    # e.g. if we see "\\t", return "\t"
    for char, escaped_mnemonic in escapes.items():
        try:
            return char, static(string, i, escaped_mnemonic)
        except NoMatch:
            pass

    # special chars e.g. "\\-" returns "-"
    for char in Charclass.classSpecial:
        try:
            return char, static(string, i, "\\" + char)
        except NoMatch:
            pass

    # hex escape e.g. "\\x40" returns "@"
    try:
        return unescape_hex(string, i)
    except NoMatch:
        pass

    # single non-special character, not contained
    # inside square brackets
    char, j = string[i], i + 1
    if char in Charclass.classSpecial:
        raise NoMatch

    return char, j


def match_inner_charclass(
    string: str,
    i: int,
) -> MatchResult[Charclass]:
    """
    We have to return several ranges, because of \\\\w etc.
    """
    # Attempt 1: shorthand
    inner_shorthand = {
        "\\w": WORDCHAR,
        "\\d": DIGIT,
        "\\s": SPACECHAR,
        "\\W": NONWORDCHAR,
        "\\D": NONDIGITCHAR,
        "\\S": NONSPACECHAR,
        # no ".": DOT,
    }

    for cc_shorthand, charclass in inner_shorthand.items():
        try:
            return charclass, static(string, i, cc_shorthand)
        except NoMatch:
            pass

    # Attempt 2: a range e.g. "d-h"
    try:
        first, j = match_internal_char(string, i)  # `first` is "d"
        k = static(string, j, "-")
        last, k = match_internal_char(string, k)  # `last` is "h"
        return Charclass(((first, last),)), k
    except NoMatch:
        pass

    # Attempt 3: just a character on its own
    char, j = match_internal_char(string, i)
    return Charclass(((char, char),)), j


def match_class_interior(string: str, i: int) -> MatchResult[Charclass]:
    inner_charclasses = []
    try:
        while True:
            # Match an internal character, range, or other charclass predicate.
            inner_charclass, i = match_inner_charclass(string, i)
            inner_charclasses.append(inner_charclass)
    except NoMatch:
        pass

    # Use the existing Charclass union functionality
    charclass = Charclass()
    for inner_charclass in inner_charclasses:
        charclass |= inner_charclass

    return charclass, i


def match_charclass(string: str, i: int) -> MatchResult[Charclass]:
    # pylint: disable=too-many-return-statements

    if i >= len(string):
        raise NoMatch

    # wildcard ".", "\\w", "\\d", etc.
    for shorthand_charclass, shorthand_abbrev in shorthand.items():
        try:
            return shorthand_charclass, static(string, i, shorthand_abbrev)
        except NoMatch:
            pass

    # "[^dsgsdg]"
    try:
        j = static(string, i, "[^")
        result, j = match_class_interior(string, j)
        j = static(string, j, "]")
        return ~result, j
    except NoMatch:
        pass

    # "[sdfsf]"
    try:
        j = static(string, i, "[")
        result, j = match_class_interior(string, j)
        j = static(string, j, "]")
        return result, j
    except NoMatch:
        pass

    # e.g. if seeing "\\t", return "\t"
    for char, escaped_mnemonic in escapes.items():
        try:
            return Charclass(((char, char),)), static(string, i, escaped_mnemonic)
        except NoMatch:
            pass

    # e.g. if seeing "\\{", return "{"
    for char in Charclass.allSpecial:
        try:
            return Charclass(((char, char),)), static(string, i, "\\" + char)
        except NoMatch:
            pass

    # e.g. if seeing "\\x40", return "@"
    try:
        char, j = unescape_hex(string, i)
        return Charclass(((char, char),)), j
    except NoMatch:
        pass

    # single non-special character, not contained inside square brackets
    char, i = string[i], i + 1
    if char in Charclass.allSpecial:
        raise NoMatch

    return Charclass(((char, char),)), i


def match_multiplicand(string: str, i: int) -> MatchResult[Pattern | Charclass]:
    # explicitly non-capturing "(?:...)" syntax. No special significance
    try:
        j = static(string, i, "(?")
        opts, j = select_static(string, j, ":", "P<")
        if opts == "P<":
            _group_name, j = read_until(string, j, ">")
        pattern, j = match_pattern(string, j)
        j = static(string, j, ")")
        return pattern, j
    except NoMatch:
        pass

    # normal "(...)" syntax
    try:
        j = static(string, i, "(")
        pattern, j = match_pattern(string, j)
        j = static(string, j, ")")
        return pattern, j
    except NoMatch:
        pass

    # Just a `Charclass` on its own
    charclass, j = match_charclass(string, i)
    return charclass, j


def match_any_of(string: str, i: int, collection: Collection[str]) -> MatchResult[str]:
    for char in collection:
        try:
            return char, static(string, i, char)
        except NoMatch:
            pass
    raise NoMatch


def match_bound(string: str, i: int) -> MatchResult[Bound]:
    # "0"
    try:
        return Bound(0), static(string, i, "0")
    except NoMatch:
        pass

    # "1", etc.
    try:
        digit, j = match_any_of(string, i, "123456789")
        integer = int(digit)
        try:
            while True:
                digit, j = match_any_of(string, j, "0123456789")
                integer *= 10
                integer += int(digit)
        except NoMatch:
            return Bound(integer), j
    except NoMatch:
        pass

    # "" empty string = infinite bound as in "{4,}"
    return INF, i


def match_nonempty_greedy_multiplier(string: str, i: int) -> MatchResult[Multiplier]:
    """
    Any multiplier which isn't the default empty string (equivalent to `{1,1}`)
    """
    # {2,3} or {2,}
    try:
        j = static(string, i, "{")
        min_, j = match_bound(string, j)
        j = static(string, j, ",")
        max_, j = match_bound(string, j)
        j = static(string, j, "}")
        return Multiplier(min_, max_), j
    except NoMatch:
        pass

    # {2}
    try:
        j = static(string, i, "{")
        min_, j = match_bound(string, j)
        j = static(string, j, "}")
        return Multiplier(min_, min_), j
    except NoMatch:
        pass

    # "?"/"*"/"+"
    for mult, symbol in symbolic.items():
        if not symbol:
            continue
        try:
            return mult, static(string, i, symbol)
        except NoMatch:
            pass

    raise NoMatch


def match_nonempty_multiplier(string: str, i: int) -> MatchResult[Multiplier]:
    multiplier, j = match_nonempty_greedy_multiplier(string, i)
    try:
        j = static(string, j, "?")
    except NoMatch:
        pass
    return multiplier, j


def match_multiplier(string: str, i: int) -> MatchResult[Multiplier]:
    try:
        return match_nonempty_multiplier(string, i)
    except NoMatch:
        return ONE, i


def match_mult(string: str, i: int) -> MatchResult[Mult]:
    multiplicand, j = match_multiplicand(string, i)
    multiplier, j = match_multiplier(string, j)
    return Mult(multiplicand, multiplier), j


def match_conc(string: str, i: int) -> MatchResult[Conc]:
    mults = []
    try:
        while True:
            m, i = match_mult(string, i)
            mults.append(m)
    except NoMatch:
        pass
    return Conc(*mults), i


def match_pattern(string: str, i: int) -> MatchResult[Pattern]:
    concs = []

    # first one
    c, i = match_conc(string, i)
    concs.append(c)

    # the rest
    while True:
        try:
            i = static(string, i, "|")
            c, i = match_conc(string, i)
            concs.append(c)
        except NoMatch:
            return Pattern(*concs), i


def parse(string: str) -> Pattern:
    """
    Parse a full string and return a `Pattern` object. Fail if
    the whole string wasn't parsed
    """
    obj, i = match_pattern(string, 0)
    if i != len(string):
        raise NoMatch(f"Could not parse {string!r} beyond index {i}")
    return obj



"""
Finite state machine library, intended to be used by `greenery` only
"""

from __future__ import annotations

__all__ = (
    "Fsm",
    "StateType",
    "EPSILON",
    "NULL",
    "Charclass",
)

from dataclasses import dataclass
from typing import (
    Callable,
    ClassVar,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    TypeVar,
)

from .charclass import DOT, Charclass, repartition

AlphaType = Charclass
StateType = int
M = TypeVar("M")
"""Meta-state type for crawl(). Can be anything."""


def unify_alphabets(fsms: Iterable[Fsm], /) -> List[Fsm]:
    charclasses = set()
    for fsm in fsms:
        for charclass in fsm.alphabet:
            charclasses.add(charclass)

    partition = repartition(charclasses)
    # maps old Charclasses to collections of new Charclasses

    return [fsm.replace_alphabet(partition) for fsm in fsms]


# pylint: disable=too-many-public-methods,too-many-branches,fixme
@dataclass(frozen=True, init=False)
class Fsm:
    """
    A Finite State Machine or FSM has an alphabet and a set of states. At
    any given moment, the FSM is in one state. When passed a symbol from
    the alphabet, the FSM jumps to another state (or possibly the same
    state). A map (Python dictionary) indicates where to jump.
    One state is nominated as a starting state. Zero or more states are
    nominated as final states. If, after consuming a string of symbols,
    the FSM is in a final state, then it is said to "accept" the string.
    This class also has some pretty powerful methods which allow FSMs to
    be concatenated, alternated between, multiplied, looped (Kleene star
    closure), intersected, and simplified.
    The majority of these methods are available using operator overloads.
    """

    alphabet: frozenset[AlphaType]
    states: frozenset[StateType]
    initial: StateType
    finals: frozenset[StateType]
    map: Mapping[StateType, Mapping[AlphaType, StateType]]

    # noinspection PyShadowingBuiltins
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        /,
        *,
        alphabet: Iterable[AlphaType],
        states: Iterable[StateType],
        initial: StateType,
        finals: Iterable[StateType],
        # pylint: disable=redefined-builtin
        map: Mapping[StateType, Mapping[AlphaType, StateType]],
    ) -> None:
        """
        `alphabet` is an iterable of symbols the FSM can be fed.
        `states` is the set of states for the FSM
        `initial` is the initial state
        `finals` is the set of accepting states
        `map` may be sparse (i.e. it may omit transitions). In the case of
        omitted transitions, a non-final "oblivion" state is simulated.
        """
        alphabet = frozenset(alphabet)
        states = frozenset(states)
        finals = frozenset(finals)

        # Validation. Thanks to immutability, this only needs to be carried out
        # once.
        if initial not in states:
            raise ValueError(f"Initial state {initial!r} must be one of {states!r}")
        if not finals.issubset(states):
            raise ValueError(f"Final states {finals!r} must be a subset of {states!r}")
        for state, state_trans in map.items():
            if state not in states:
                raise ValueError(f"Transition from unknown state {state!r}")
            for symbol, dest in state_trans.items():
                if symbol not in alphabet:
                    raise ValueError(
                        f"Invalid symbol {symbol!r}"
                        f" in transition from {state!r}"
                        f" to {dest!r}"
                    )
                if dest not in states:
                    raise ValueError(
                        f"Transition for state {state!r}"
                        f" and symbol {symbol!r}"
                        f" leads to {dest!r},"
                        " which is not a state"
                    )
        for state in states:
            if state not in map:
                raise ValueError(f"State {state!r} missing from map")
            for charclass in alphabet:
                if charclass not in map[state]:
                    raise ValueError(
                        f"Symbol {charclass!r} missing from map[{state!r}]"
                    )

        # Check that the charclasses form a proper partition of all of Unicode
        unified = Charclass()
        for charclass in alphabet:
            if unified & charclass != Charclass():
                raise ValueError(f"Alphabet {alphabet!r} has overlaps")
            unified |= charclass
        if unified != DOT:
            raise ValueError(f"Alphabet {alphabet!r} is not a proper partition")

        # Initialise the hard way due to immutability.
        object.__setattr__(self, "alphabet", alphabet)
        object.__setattr__(self, "states", states)
        object.__setattr__(self, "initial", initial)
        object.__setattr__(self, "finals", finals)
        object.__setattr__(self, "map", map)

    def accepts(self, string: str, /) -> bool:
        """
        Test whether the present FSM accepts the supplied string (iterable
        of symbols). Equivalently, consider `self` as a possibly-infinite
        set of strings and test whether `string` is a member of it. This is
        actually mainly used for unit testing purposes.
        """
        state = self.initial
        for char in string:
            for charclass in self.map[state]:
                if charclass.accepts(char):
                    state = self.map[state][charclass]
                    break
        return state in self.finals

    def __contains__(self, string: str, /) -> bool:
        """
        This lets you use the syntax `"a" in fsm1` to see whether the
        string "a" is in the set of strings accepted by `fsm1`.
        """
        return self.accepts(string)

    def reduce(self, /) -> Fsm:
        """
        A result by Brzozowski (1963) shows that a minimal finite state
        machine equivalent to the original can be obtained by reversing the
        original twice.
        """
        return self.reversed().reversed()

    def __repr__(self, /) -> str:
        args = ", ".join(
            [
                f"alphabet={self.alphabet!r}",
                f"states={self.states!r}",
                f"initial={self.initial!r}",
                f"finals={self.finals!r}",
                f"map={self.map!r}",
            ]
        )
        return f"Fsm({args})"

    # The Python `__eq__` + `__hash__` contract requires that value-equality
    # implies hash-equality. `Fsm` `__eq__` implementation currently represents
    # equality of the set of accepted strings, independent of specific state
    # labels or unused members of the alphabet. This is not trivial to hash.
    # Regarding the type suppression, see
    # https://github.com/python/mypy/issues/4266
    __hash__: ClassVar[None] = None  # type: ignore

    def __str__(self, /) -> str:
        rows = []

        sorted_alphabet = sorted(self.alphabet)

        # top row
        row = ["", "name", "final?"]
        row.extend(str(symbol) for symbol in sorted_alphabet)
        rows.append(row)

        # other rows
        for state in self.states:
            row = []
            if state == self.initial:
                row.append("*")
            else:
                row.append("")
            row.append(str(state))
            if state in self.finals:
                row.append("True")
            else:
                row.append("False")
            for symbol in sorted_alphabet:
                row.append(str(self.map[state][symbol]))
            rows.append(row)

        # column widths
        colwidths = []
        for x in range(len(rows[0])):
            colwidths.append(max(len(str(row[x])) for y, row in enumerate(rows)) + 1)

        # apply padding
        for y, row in enumerate(rows):
            for x, col in enumerate(row):
                rows[y][x] = col.ljust(colwidths[x])

        # horizontal line
        rows.insert(1, ["-" * colwidth for colwidth in colwidths])

        return "".join("".join(row) + "\n" for row in rows)

    def concatenate(*fsms: Fsm) -> Fsm:
        """
        Concatenate arbitrarily many finite state machines together.
        """
        unified_fsms = unify_alphabets(fsms)

        def connect_all(
            i: int,
            substate: StateType,
        ) -> Iterable[tuple[int, StateType]]:
            """
            Take a state in the numbered FSM and return a set containing
            it, plus (if it's final) the first state from the next FSM,
            plus (if that's final) the first state from the next but one
            FSM, plus...
            """
            result = {(i, substate)}
            while i < len(unified_fsms) - 1 and substate in unified_fsms[i].finals:
                i += 1
                substate = unified_fsms[i].initial
                result.add((i, substate))
            return result

        # Use a superset containing states from all FSMs at once.
        # We start at the start of the first FSM. If this state is final in the
        # first FSM, then we are also at the start of the second FSM. And so
        # on.
        initial = frozenset(
            connect_all(0, unified_fsms[0].initial) if unified_fsms else ()
        )

        def final(state: frozenset[tuple[int, StateType]]) -> bool:
            """If you're in a final state of the final FSM, it's final"""
            return any(
                i == len(unified_fsms) - 1 and substate in unified_fsms[i].finals
                for i, substate in state
            )

        def follow(
            current: frozenset[tuple[int, StateType]],
            symbol: AlphaType,
        ) -> frozenset[tuple[int, StateType]]:
            """
            Follow the collection of states through all FSMs at once,
            jumping to the next FSM if we reach the end of the current one
            """
            next_metastate: set[tuple[int, StateType]] = set()
            for i, substate in current:
                next_metastate.update(
                    connect_all(i, unified_fsms[i].map[substate][symbol])
                )

            return frozenset(next_metastate)

        alphabet = unified_fsms[0].alphabet if len(unified_fsms) > 0 else {~Charclass()}

        return crawl(alphabet, initial, final, follow).reduce()

    def __add__(self, other: Fsm, /) -> Fsm:
        """
        Concatenate two finite state machines together.
        For example, if self accepts "0*" and other accepts "1+(0|1)",
        will return a finite state machine accepting "0*1+(0|1)".
        Accomplished by effectively following non-deterministically.
        Call using "fsm3 = fsm1 + fsm2"
        """
        return self.concatenate(other)

    def star(self, /) -> Fsm:
        """
        If the present FSM accepts X, returns an FSM accepting X* (i.e. 0
        or more Xes). This is NOT as simple as naively connecting the final
        states back to the initial state: see (b*ab)* for example.
        """
        alphabet = self.alphabet

        initial: Collection[StateType] = {self.initial}

        def follow(
            state: Collection[StateType],
            symbol: AlphaType,
        ) -> Collection[StateType]:
            next_states = set()

            for substate in state:
                next_states.add(self.map[substate][symbol])

                # If one of our substates is final, then we can also consider
                # transitions from the initial state of the original FSM.
                if substate in self.finals:
                    next_states.add(self.map[self.initial][symbol])

            return frozenset(next_states)

        def final(state: Collection[StateType]) -> bool:
            return any(substate in self.finals for substate in state)

        return crawl(alphabet, initial, final, follow) | EPSILON

    def times(self, multiplier: int, /) -> Fsm:
        """
        Given an FSM and a multiplier, return the multiplied FSM.
        """
        if multiplier < 0:
            raise ArithmeticError(f"Can't multiply an FSM by {multiplier!r}")

        alphabet = self.alphabet

        # metastate is a set of iterations+states
        initial: Collection[tuple[StateType, int]] = {(self.initial, 0)}

        def final(state: Collection[tuple[StateType, int]]) -> bool:
            """
            If the initial state is final then multiplying doesn't alter
            that
            """
            return any(
                substate == self.initial
                and (self.initial in self.finals or iteration == multiplier)
                for substate, iteration in state
            )

        def follow(
            current: Collection[tuple[StateType, int]],
            symbol: AlphaType,
        ) -> Collection[tuple[StateType, int]]:
            next_metastate = []
            for substate, iteration in current:
                if iteration < multiplier:
                    next_metastate.append((self.map[substate][symbol], iteration))
                    # final of self? merge with initial on next iteration
                    if self.map[substate][symbol] in self.finals:
                        next_metastate.append((self.initial, iteration + 1))
            return frozenset(next_metastate)

        return crawl(alphabet, initial, final, follow).reduce()

    def __mul__(self, multiplier: int, /) -> Fsm:
        """
        Given an FSM and a multiplier, return the multiplied FSM.
        """
        return self.times(multiplier)

    def union(*fsms: Fsm) -> Fsm:
        """
        Treat `fsms` as a collection of arbitrary FSMs and return the union
        FSM. Can be used as `fsm1.union(fsm2, ...)` or
        `fsm.union(fsm1, ...)`. `fsms` may be empty.
        """
        return parallel(fsms, any)

    def __or__(self, other: Fsm, /) -> Fsm:
        """
        Alternation.
        Return a finite state machine which accepts any sequence of symbols
        that is accepted by either self or other. Note that the set of
        strings recognised by the two FSMs undergoes a set union.
        Call using "fsm3 = fsm1 | fsm2"
        """
        return self.union(other)

    def intersection(*fsms: Fsm) -> Fsm:
        """
        Intersection.
        Take FSMs and AND them together. That is, return an FSM which
        accepts any sequence of symbols that is accepted by both of the
        original FSMs. Note that the set of strings recognised by the two
        FSMs undergoes a set intersection operation.
        Call using "fsm3 = fsm1 & fsm2"
        """
        return parallel(fsms, all)

    def __and__(self, other: Fsm, /) -> Fsm:
        """
        Treat the FSMs as sets of strings and return the intersection of
        those sets in the form of a new FSM.
        """
        return self.intersection(other)

    def symmetric_difference(*fsms: Fsm) -> Fsm:
        """
        Treat `fsms` as a collection of sets of strings and compute the
        symmetric difference of them all. The python set method only allows
        two sets to be operated on at once, but we go the extra mile since
        it's not too hard.
        """
        return parallel(fsms, lambda accepts: (accepts.count(True) % 2) == 1)

    def __xor__(self, other: Fsm, /) -> Fsm:
        """
        Symmetric difference. Returns an FSM which recognises only the
        strings recognised by `self` or `other` but not both.
        """
        return self.symmetric_difference(other)

    def everythingbut(self, /) -> Fsm:
        """
        Return a finite state machine which will accept any string NOT
        accepted by self, and will not accept any string accepted by self.
        """
        alphabet = self.alphabet
        initial = self.initial

        def follow(
            current: StateType,
            symbol: AlphaType,
        ) -> StateType:
            return self.map[current][symbol]

        # state is final unless the original was
        def final(state: StateType) -> bool:
            return state not in self.finals

        return crawl(alphabet, initial, final, follow).reduce()

    def reversed(self, /) -> Fsm:
        """
        Return a new FSM such that for every string that self accepts (e.g.
        "beer", the new FSM accepts the reversed string ("reeb").
        """
        alphabet = self.alphabet

        # Start from a composite "state-set" consisting of all final states.
        # If there are no final states, this set is empty and we'll find that
        # no other states get generated.
        initial = frozenset(self.finals)

        # Find every possible way to reach the current state-set
        # using this symbol.
        def follow(
            current: frozenset[StateType],
            symbol: AlphaType,
        ) -> frozenset[StateType]:
            next_states = frozenset(
                [
                    prev
                    for prev in self.map
                    for state in current
                    if self.map[prev][symbol] == state
                ]
            )
            return next_states

        # A state-set is final if the initial state is in it.
        def final(state: frozenset[StateType]) -> bool:
            return self.initial in state

        # Man, crawl() is the best!
        return crawl(alphabet, initial, final, follow)
        # Do not reduce() the result, since reduce() calls us in turn

    def islive(self, /, state: StateType) -> bool:
        """A state is "live" if a final state can be reached from it."""
        reachable = [state]
        i = 0
        while i < len(reachable):
            current = reachable[i]
            if current in self.finals:
                return True
            for symbol in self.map[current]:
                next_state = self.map[current][symbol]
                if next_state not in reachable:
                    reachable.append(next_state)
            i += 1
        return False

    def empty(self, /) -> bool:
        """
        An FSM is empty if it recognises no strings. An FSM may be
        arbitrarily complicated and have arbitrarily many final states
        while still recognising no strings because those final states may
        all be inaccessible from the initial state. Equally, an FSM may be
        non-empty despite having an empty alphabet if the initial state is
        final.
        """
        return not self.islive(self.initial)

    def strings(self, otherchars: Iterable[str]) -> Iterator[str]:
        """
        Generate strings that this FSM accepts. Note that for our purposes a
        string is a sequence of Unicode characters, NOT a list of Charclasses.

        Since
        there may be infinitely many of these we use a generator instead of
        constructing a static list. Strings will be sorted in order of
        length and then lexically. This procedure uses arbitrary amounts of
        memory but is very fast. There may be more efficient ways to do
        this, that I haven't investigated yet. You can use this in list
        comprehensions.
        """

        # Most FSMs have at least one "dead state".
        # Once you reach a dead state, you can no
        # longer reach a final state. Since many strings may end up here, it's
        # advantageous to constrain our search to live states only.
        livestates = set(state for state in self.states if self.islive(state))

        # We store a list of tuples. Each tuple consists of an input string and
        # the state that this input string leads to. This means we don't have
        # to run the state machine from the very beginning every time we want
        # to check a new string.
        strings: list[tuple[str, StateType]] = []

        # Initial entry (or possibly not, in which case this is a short one)
        cstate: StateType = self.initial
        cstring: str = ""
        if cstate in livestates:
            if cstate in self.finals:
                yield cstring
            strings.append((cstring, cstate))

        # Fixed point calculation
        i = 0
        while i < len(strings):
            cstring, cstate = strings[i]

            for charclass in sorted(self.map[cstate]):
                # TODO: scrap otherchars as a concept?
                chars = otherchars if charclass.negated else charclass.get_chars()
                for char in chars:
                    nstate = self.map[cstate][charclass]
                    nstring = cstring + char
                    if nstate in livestates:
                        if nstate in self.finals:
                            yield nstring
                        strings.append((nstring, nstate))
            i += 1

    def __iter__(self, /) -> Iterator[str]:
        """
        This allows you to do `for string in fsm1` as a list comprehension!
        """
        return self.strings([])

    def equivalent(self, other: Fsm, /) -> bool:
        """
        Two FSMs are considered equivalent if they recognise the same
        strings. Or, to put it another way, if their symmetric difference
        recognises no strings.
        """
        return (self ^ other).empty()

    def __eq__(self, other: object, /) -> bool:
        """
        You can use `fsm1 == fsm2` to determine whether two FSMs recognise
        the same strings.
        """
        if not isinstance(other, Fsm):
            return NotImplemented
        return self.equivalent(other)

    def different(self, other: Fsm, /) -> bool:
        """
        Two FSMs are considered different if they have a non-empty
        symmetric difference.
        """
        return not (self ^ other).empty()

    def __ne__(self, other: object, /) -> bool:
        """
        Use `fsm1 != fsm2` to determine whether two FSMs recognise
        different strings.
        """
        return not self == other

    def difference(*fsms: Fsm) -> Fsm:
        """
        Difference. Returns an FSM which recognises only the strings
        recognised by the first FSM in the list, but none of the others.
        """
        return parallel(fsms, lambda accepts: accepts[0] and not any(accepts[1:]))

    def __sub__(self, other: Fsm, /) -> Fsm:
        return self.difference(other)

    def cardinality(self, /) -> int:
        """
        Consider the FSM as a set of strings and return the cardinality of
        that set, or raise an OverflowError if there are infinitely many
        """
        num_strings: dict[StateType, int | None] = {}

        def get_num_strings(state: StateType) -> int:
            # Most FSMs have at least one oblivion state
            if self.islive(state):
                if state in num_strings:
                    if num_strings[state] is None:  # "computing..."
                        # Recursion! There are infinitely many strings
                        # recognised
                        raise OverflowError(state)
                    return num_strings[state]  # type: ignore

                num_strings[state] = None  # i.e. "computing..."
                n = 0
                for charclass in self.map[state]:
                    num_transitions = charclass.num_chars()
                    nstate = self.map[state][charclass]
                    if nstate in self.finals:
                        n += num_transitions
                    n += num_transitions * get_num_strings(nstate)
                num_strings[state] = n

            else:
                # Dead state
                num_strings[state] = 0

            return num_strings[state]  # type: ignore

        n = 1 if self.initial in self.finals else 0
        return n + get_num_strings(self.initial)

    def __len__(self, /) -> int:
        """
        Consider the FSM as a set of strings and return the cardinality of
        that set, or raise an OverflowError if there are infinitely many
        """
        return self.cardinality()

    def isdisjoint(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if they are
        disjoint
        """
        return (self & other).empty()

    def issubset(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        subset of `other`... `self` recognises no strings which `other`
        doesn't.
        """
        return (self - other).empty()

    def __le__(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        subset of `other`... `self` recognises no strings which `other`
        doesn't.
        """
        return self.issubset(other)

    def ispropersubset(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        proper subset of `other`.
        """
        return self <= other and self != other

    def __lt__(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        strict subset of `other`.
        """
        return self.ispropersubset(other)

    def issuperset(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        superset of `other`.
        """
        return (other - self).empty()

    def __ge__(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        superset of `other`.
        """
        return self.issuperset(other)

    def ispropersuperset(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        proper superset of `other`.
        """
        return self >= other and self != other

    def __gt__(self, other: Fsm, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if `self` is a
        strict superset of `other`.
        """
        return self.ispropersuperset(other)

    def copy(self, /) -> Fsm:
        """
        For completeness only, since `set.copy()` and `frozenset.copy()` exist.
        FSM objects are immutable; like `frozenset`, this just returns `self`.
        """
        return self

    __copy__ = copy

    def derive(self, string: str, /) -> Fsm:
        """
        Compute the Brzozowski derivative of this FSM with respect to the
        input string. Note that the FSM uses Charclasses as symbols internally,
        but the input string is a sequence of Unicode characters
        <https://en.wikipedia.org/wiki/Brzozowski_derivative>
        """
        # Consume the input string.
        state = self.initial
        for char in string:
            for charclass in self.map[state]:
                if charclass.accepts(char):
                    state = self.map[state][charclass]
                    break

        # OK so now we have consumed that string, use the new location as
        # the starting point.
        return Fsm(
            alphabet=self.alphabet,
            states=self.states,
            initial=state,
            finals=self.finals,
            map=self.map,
        )

    def replace_alphabet(
        self, replacements: Mapping[AlphaType, Iterable[AlphaType]]
    ) -> Fsm:
        """
        Returns a new FSM which uses a different alphabet. If one original
        symbol converts to two new symbols, there will be multiple identical
        transitions; if none, the transitions will be omitted.
        """
        new_alphabet = set()
        for symbol in self.alphabet:
            for replacement in replacements[symbol]:
                new_alphabet.add(replacement)

        new_map: Dict[StateType, Dict[AlphaType, StateType]] = {}
        for state in self.map:
            new_map[state] = {}
            for symbol in self.alphabet:
                for replacement in replacements[symbol]:
                    new_map[state][replacement] = self.map[state][symbol]

        return Fsm(
            alphabet=new_alphabet,
            states=self.states,
            initial=self.initial,
            finals=self.finals,
            map=new_map,
        )


NULL = Fsm(
    alphabet={~Charclass()},
    states={0},
    initial=0,
    finals=(),
    map={
        0: {~Charclass(): 0},
    },
)
"""
An FSM accepting nothing (not even the empty string). This is
demonstrates that this is possible, and is also extremely useful
in some situations
"""

EPSILON = Fsm(
    alphabet={~Charclass()},
    states={0, 1},
    initial=0,
    finals={0},
    map={
        0: {~Charclass(): 1},
        1: {~Charclass(): 1},
    },
)
"""
An FSM matching an empty string, "", only.
This is very useful in many situations
"""


def parallel(
    fsms: tuple[Fsm, ...],
    test: Callable[[list[bool]], bool],
    /,
) -> Fsm:
    """
    Crawl several FSMs in parallel, mapping the states of a larger
    meta-FSM. To determine whether a state in the larger FSM is final, pass
    all of the finality statuses (e.g. [True, False, False] to `test`.
    """
    unified_fsms = unify_alphabets(fsms)

    initial: Mapping[int, StateType] = {
        i: fsm.initial for i, fsm in enumerate(unified_fsms)
    }

    # dedicated function accepts a "superset" and returns the next "superset"
    # obtained by following this transition in the new FSM
    def follow(
        current: Mapping[int, StateType],
        symbol: AlphaType,
    ) -> Mapping[int, StateType]:
        return {i: fsm.map[current[i]][symbol] for i, fsm in enumerate(unified_fsms)}

    # Determine the "is final?" condition of each substate, then pass it to the
    # test to determine finality of the overall FSM.
    def final(state: Mapping[int, StateType]) -> bool:
        return test([state[i] in fsm.finals for i, fsm in enumerate(unified_fsms)])

    alphabet = unified_fsms[0].alphabet if len(unified_fsms) > 0 else {~Charclass()}

    return crawl(alphabet, initial, final, follow).reduce()


def crawl(
    alphabet: Iterable[AlphaType],
    initial: M,
    final: Callable[[M], bool],
    follow: Callable[[M, AlphaType], M],
) -> Fsm:
    """
    Given the above conditions and instructions, crawl a new unknown FSM,
    mapping its states, final states and transitions. Return the new FSM.
    This is a pretty powerful procedure which could potentially go on
    forever if you supply an evil version of follow().
    """

    states: list[M] = [initial]
    finals: set[StateType] = set()
    transitions: dict[StateType, dict[AlphaType, StateType]] = {}

    # iterate over a growing list
    i = 0
    while i < len(states):
        state = states[i]

        # add to finals
        if final(state):
            finals.add(i)

        # compute map for this state
        transitions[i] = {}
        for symbol in sorted(alphabet):
            next_state = follow(state, symbol)

            try:
                j = states.index(next_state)
            except ValueError:
                j = len(states)
                states.append(next_state)

            transitions[i][symbol] = j

        i += 1

    return Fsm(
        alphabet=alphabet,
        states=set(range(len(states))),
        initial=0,
        finals=finals,
        map=transitions,
    )


def from_charclass(charclass: Charclass) -> Fsm:
    # 0 is initial, 1 is final, 2 is dead
    return Fsm(
        alphabet={charclass, ~charclass},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {charclass: 1, ~charclass: 2},
            1: {charclass: 2, ~charclass: 2},
            2: {charclass: 2, ~charclass: 2},
        },
    )



from __future__ import annotations

from .charclass import Charclass
from .multiplier import ONE, ZERO
from .parse import parse
from .rxelems import Conc, Mult, Pattern


def test_pattern_equality() -> None:
    assert Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("b"), ONE)),
    ) == Pattern(
        Conc(Mult(Charclass("b"), ONE)),
        Conc(Mult(Charclass("a"), ONE)),
    )
    assert Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("a"), ONE)),
    ) == Pattern(
        Conc(Mult(Charclass("a"), ONE)),
    )


def test_pattern_str() -> None:
    assert (
        str(
            Pattern(
                Conc(Mult(Charclass("a"), ONE)),
                Conc(Mult(Charclass("b"), ONE)),
            )
        )
        == "a|b"
    )
    assert (
        str(
            Pattern(
                Conc(Mult(Charclass("a"), ONE)),
                Conc(Mult(Charclass("a"), ONE)),
            )
        )
        == "a"
    )
    assert (
        str(
            Pattern(
                Conc(
                    Mult(Charclass("a"), ONE),
                    Mult(Charclass("b"), ONE),
                    Mult(Charclass("c"), ONE),
                ),
                Conc(
                    Mult(Charclass("d"), ONE),
                    Mult(Charclass("e"), ONE),
                    Mult(Charclass("f"), ONE),
                    Mult(
                        Pattern(
                            Conc(
                                Mult(Charclass("g"), ONE),
                                Mult(Charclass("h"), ONE),
                                Mult(Charclass("i"), ONE),
                            ),
                            Conc(
                                Mult(Charclass("j"), ONE),
                                Mult(Charclass("k"), ONE),
                                Mult(Charclass("l"), ONE),
                            ),
                        ),
                        ONE,
                    ),
                ),
            )
        )
        == "abc|def(ghi|jkl)"
    )


def test_empty() -> None:
    assert Pattern().empty()


def test_mult_reduction_easy() -> None:
    assert Pattern(Conc()).reduce() == Pattern(Conc())
    assert Pattern(
        Conc(
            Mult(
                Charclass("a"),
                ZERO,
            )
        )
    ).reduce() == Pattern(Conc())

    assert str(
        # pylint: disable-next=compare-to-empty-string
        Pattern(
            Conc(
                Mult(
                    Charclass("a"),
                    ZERO,
                )
            ).reduce()
        )
        == ""
    )


def test_empty_pattern_reduction() -> None:
    assert str(Pattern().reduce()) == "[]"


def test_empty_conc_suppression() -> None:
    assert (
        str(
            Pattern(
                Conc(
                    # this `Mult` can never actually match anything
                    Mult(Pattern(), ONE),
                    Mult(Charclass("0"), ONE),
                    Mult(Charclass("0123456789"), ONE),
                )  # so neither can this `Conc`
            ).reduce()
        )
        == "[]"
    )


def test_pattern_dock() -> None:
    a = Mult(Charclass("a"), ONE)
    c = Mult(Charclass("c"), ONE)
    f = Mult(Charclass("f"), ONE)

    assert parse("a|bc").dock(Conc()) == parse("a|bc")
    assert parse("aa|bca").dock(Conc(a)) == parse("a|bc")
    assert parse("xyza|abca|a").dock(Conc(a)) == parse("xyz|abc|")
    assert parse("f{2,3}c|fc").dock(Conc(f, c)) == parse("f{1,2}|")
    assert parse("aa").dock(Conc(a, a)) == parse("")


def test_pattern_beheading() -> None:
    a = Mult(Charclass("a"), ONE)
    c = Mult(Charclass("c"), ONE)
    f = Mult(Charclass("f"), ONE)
    z = Mult(Charclass("Z"), ONE)

    assert parse("aa").behead(Conc(a)) == parse("a")
    assert parse("abc|aa").behead(Conc(a)) == parse("a|bc")
    assert parse("cf{1,2}|cf").behead(Conc(c)) == parse("f{1,2}|f")
    assert parse("aa|aa").behead(Conc(a, a)) == parse("")
    assert parse("abc|aa").behead(Conc(a)) == parse("a|bc")
    assert parse("a|bc").behead(Conc()) == parse("a|bc")
    assert parse("cf{1,2}|cf").behead(Conc(c, f)) == parse("f?|")
    assert parse("ZA|ZB|ZC").behead(Conc(z)) == parse("A|B|C")
    assert parse("Z+A|ZB|ZZC").behead(Conc(z)) == parse("Z*A|B|ZC")
    assert parse("a{2}b|a+c").behead(Conc(a)) == parse("ab|a*c")



from __future__ import annotations

import pytest

from .bound import INF, Bound
from .multiplier import ONE, PLUS, QM, STAR, ZERO, Multiplier


def test_multiplier_str() -> None:
    assert str(Multiplier(Bound(2), INF)) == "{2,}"
    assert str(Multiplier(Bound(0), Bound(0))) == "{0}"
    assert str(Multiplier(Bound(2), Bound(2))) == "{2}"
    assert str(Multiplier(Bound(2), Bound(5))) == "{2,5}"


def test_bound_qm() -> None:
    assert QM.mandatory == Bound(0)
    assert QM.optional == Bound(1)


def test_eq() -> None:
    assert ZERO == Multiplier(Bound(0), Bound(0))
    assert ONE == Multiplier(Bound(1), Bound(1))
    assert STAR == Multiplier(Bound(0), INF)
    assert Multiplier(Bound(1), Bound(2)) == Multiplier(Bound(1), Bound(2))

    assert ZERO != ONE
    assert STAR != QM


def test_eq_het() -> None:
    assert ZERO != "goldfish"


def test_multiplier_common() -> None:
    assert ZERO.common(ZERO) == ZERO
    assert ZERO.common(QM) == ZERO
    assert ZERO.common(ONE) == ZERO
    assert ZERO.common(STAR) == ZERO
    assert ZERO.common(PLUS) == ZERO
    assert QM.common(ZERO) == ZERO
    assert QM.common(QM) == QM
    assert QM.common(ONE) == ZERO
    assert QM.common(STAR) == QM
    assert QM.common(PLUS) == QM
    assert ONE.common(ZERO) == ZERO
    assert ONE.common(QM) == ZERO
    assert ONE.common(ONE) == ONE
    assert ONE.common(STAR) == ZERO
    assert ONE.common(PLUS) == ONE
    assert STAR.common(ZERO) == ZERO
    assert STAR.common(QM) == QM
    assert STAR.common(ONE) == ZERO
    assert STAR.common(STAR) == STAR
    assert STAR.common(PLUS) == STAR
    assert PLUS.common(ZERO) == ZERO
    assert PLUS.common(QM) == QM
    assert PLUS.common(ONE) == ONE
    assert PLUS.common(STAR) == STAR
    assert PLUS.common(PLUS) == PLUS


def test_multiplier_subtraction() -> None:
    # a{3,4}, a{2,5} -> a{2,3} (with a{1,1}, a{0,2} left over)
    assert Multiplier(Bound(3), Bound(4)).common(
        Multiplier(Bound(2), Bound(5))
    ) == Multiplier(Bound(2), Bound(3))
    assert Multiplier(Bound(3), Bound(4)) - Multiplier(Bound(2), Bound(3)) == ONE
    assert Multiplier(Bound(2), Bound(5)) - Multiplier(
        Bound(2), Bound(3)
    ) == Multiplier(Bound(0), Bound(2))

    # a{2,}, a{1,5} -> a{1,5} (with a{1,}, a{0,0} left over)
    assert Multiplier(Bound(2), INF).common(
        Multiplier(Bound(1), Bound(5))
    ) == Multiplier(Bound(1), Bound(5))
    assert Multiplier(Bound(2), INF) - Multiplier(Bound(1), Bound(5)) == PLUS
    assert Multiplier(Bound(1), Bound(5)) - Multiplier(Bound(1), Bound(5)) == ZERO

    # a{3,}, a{2,} -> a{2,} (with a, epsilon left over)
    assert Multiplier(Bound(3), INF).common(Multiplier(Bound(2), INF)) == Multiplier(
        Bound(2), INF
    )
    assert Multiplier(Bound(3), INF) - Multiplier(Bound(2), INF) == ONE
    assert Multiplier(Bound(2), INF) - Multiplier(Bound(2), INF) == ZERO

    # a{3,}, a{3,} -> a{3,} (with ZERO, ZERO left over)
    assert Multiplier(Bound(3), INF).common(Multiplier(Bound(3), INF)) == Multiplier(
        Bound(3), INF
    )
    assert Multiplier(Bound(3), INF) - Multiplier(Bound(3), INF) == ZERO


def test_multiplier_union() -> None:
    assert ZERO | ZERO == ZERO
    assert ZERO | QM == QM
    assert ZERO | ONE == QM
    assert ZERO | STAR == STAR
    assert ZERO | PLUS == STAR
    assert QM | ZERO == QM
    assert QM | QM == QM
    assert QM | ONE == QM
    assert QM | STAR == STAR
    assert QM | PLUS == STAR
    assert ONE | ZERO == QM
    assert ONE | QM == QM
    assert ONE | ONE == ONE
    assert ONE | STAR == STAR
    assert ONE | PLUS == PLUS
    assert STAR | ZERO == STAR
    assert STAR | QM == STAR
    assert STAR | ONE == STAR
    assert STAR | STAR == STAR
    assert STAR | PLUS == STAR
    assert PLUS | ZERO == STAR
    assert PLUS | QM == STAR
    assert PLUS | ONE == PLUS
    assert PLUS | STAR == STAR
    assert PLUS | PLUS == PLUS
    assert not ZERO.canunion(Multiplier(Bound(2), INF))
    assert not ONE.canunion(Multiplier(Bound(3), Bound(4)))
    assert not Multiplier(Bound(8), INF).canunion(Multiplier(Bound(3), Bound(4)))

    with pytest.raises(ArithmeticError, match="Can't compute the union"):
        _ = ZERO | Multiplier(Bound(7), Bound(8))



from __future__ import annotations

from setuptools import setup

setup(
    name="greenery",
    version="4.2.2",
    tests_require=["pytest"],
    packages=["greenery"],
    package_dir={"greenery": "greenery"},
    author="qntm",
    author_email="qntm <qntm@users.noreply.github.com>",
    description="Greenery allows manipulation of regular expressions",
    license="MIT License",
    keywords=" ".join(
        [
            "re",
            "regex",
            "regexp",
            "regular",
            "expression",
            "deterministic",
            "finite",
            "state",
            "machine",
            "automaton",
            "fsm",
            "dfsm",
            "fsa",
            "dfsa",
            "greenery",
        ]
    ),
    url="https://github.com/qntm/greenery",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
)



from __future__ import annotations

from .bound import INF, Bound
from .charclass import DIGIT, Charclass
from .multiplier import ONE, PLUS, QM, STAR, Multiplier
from .rxelems import Mult


def test_mult_equality() -> None:
    a = Mult(Charclass("a"), ONE)
    # pylint: disable=comparison-with-itself
    assert a == a
    assert a != Mult(Charclass("b"), ONE)
    assert a != Mult(Charclass("a"), QM)
    assert a != Mult(Charclass("a"), Multiplier(Bound(1), Bound(2)))


def test_mult_str() -> None:
    a = Charclass("a")
    assert str(Mult(a, ONE)) == "a"
    assert str(Mult(a, Multiplier(Bound(2), Bound(2)))) == "a{2}"
    assert str(Mult(a, Multiplier(Bound(3), Bound(3)))) == "a{3}"
    assert str(Mult(a, Multiplier(Bound(4), Bound(4)))) == "a{4}"
    assert str(Mult(a, Multiplier(Bound(5), Bound(5)))) == "a{5}"
    assert str(Mult(a, QM)) == "a?"
    assert str(Mult(a, STAR)) == "a*"
    assert str(Mult(a, PLUS)) == "a+"
    assert str(Mult(a, Multiplier(Bound(2), Bound(5)))) == "a{2,5}"
    assert str(Mult(a, Multiplier(Bound(2), INF))) == "a{2,}"

    assert str(Mult(DIGIT, ONE)) == "\\d"
    assert str(Mult(DIGIT, Multiplier(Bound(2), Bound(2)))) == "\\d{2}"
    assert str(Mult(DIGIT, Multiplier(Bound(3), Bound(3)))) == "\\d{3}"


def test_odd_bug() -> None:
    # pylint: disable=invalid-name

    # Odd bug with ([bc]*c)?[ab]*
    int5A = Mult(
        Charclass("bc"),
        STAR,
    ).to_fsm()
    assert int5A.accepts("")

    int5B = Mult(
        Charclass("c"),
        ONE,
    ).to_fsm()
    assert int5B.accepts("c")

    int5C = int5A.concatenate(int5B)
    assert int5C.accepts("c")


def test_mult_common() -> None:
    a = Charclass("a")
    assert Mult(a, Multiplier(Bound(3), Bound(4))).common(
        Mult(a, Multiplier(Bound(2), Bound(5)))
    ) == Mult(a, Multiplier(Bound(2), Bound(3)))
    assert Mult(a, Multiplier(Bound(2), INF)).common(
        Mult(a, Multiplier(Bound(1), Bound(5)))
    ) == Mult(a, Multiplier(Bound(1), Bound(5)))
    assert Mult(a, Multiplier(Bound(3), INF)).common(
        Mult(a, Multiplier(Bound(2), INF))
    ) == Mult(a, Multiplier(Bound(2), INF))


def test_mult_dock() -> None:
    a = Charclass("a")
    assert Mult(a, Multiplier(Bound(4), Bound(5))).dock(
        Mult(a, Multiplier(Bound(3), Bound(3)))
    ) == Mult(a, Multiplier(Bound(1), Bound(2)))



from __future__ import annotations

__all__ = (
    "Multiplier",
    "ONE",
    "PLUS",
    "QM",
    "STAR",
    "ZERO",
    "symbolic",
)

from dataclasses import dataclass, field
from typing import Mapping

from .bound import INF, Bound


@dataclass(frozen=True)
class Multiplier:
    """
    A min and a max. The vast majority of characters in regular expressions
    occur without a specific multiplier, which is implicitly equivalent to
    a min of 1 and a max of 1, but many more have explicit multipliers like
    "*" (min = 0, max = inf) and so on.

    Although it seems odd and can lead to some confusing edge cases, we do
    also permit a max of 0 (iff min is 0 too). This allows the multiplier
    `ZERO` to exist, which actually are quite useful in their own special
    way.
    """

    min: Bound
    max: Bound
    mandatory: Bound = field(init=False)
    optional: Bound = field(init=False)

    def __post_init__(self, /) -> None:
        if self.min == INF:
            raise ValueError(f"Minimum bound of a multiplier can't be {INF!r}")
        if self.min > self.max:
            raise ValueError(
                f"Invalid multiplier bounds: {self.min!r} and {self.max!r}"
            )

        # More useful than "min" and "max" in many situations
        # are "mandatory" and "optional".
        object.__setattr__(self, "mandatory", self.min)
        object.__setattr__(self, "optional", self.max - self.min)

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.min == other.min and self.max == other.max

    def __hash__(self, /) -> int:
        return hash((self.min, self.max))

    def __repr__(self, /) -> str:
        return f"Multiplier({self.min!r}, {self.max!r})"

    def __str__(self, /) -> str:
        try:
            return symbolic[self]
        except LookupError:
            pass

        if self.min == self.max:
            return "{" + str(self.min) + "}"
        return "{" + str(self.min) + "," + str(self.max) + "}"

    def canmultiplyby(self, other: Multiplier, /) -> bool:
        """
        Multiplication is not well-defined for all pairs of multipliers
        because the resulting possibilities do not necessarily form a
        continuous range.

        For example:
            {0,x} * {0,y} = {0,x*y}
            {2} * {3} = {6}
            {2} * {1,2} = ERROR

        The proof isn't simple but suffice it to say that {p,p+q} * {r,r+s}
        is equal to {pr, (p+q)(r+s)} only if s=0 or qr+1 >= p. If not, then
        at least one gap appears in the range. The first inaccessible
        number is (p+q)r+1. And no, multiplication is not commutative
        """
        return (
            other.optional == Bound(0)
            or self.optional * other.mandatory + Bound(1) >= self.mandatory
        )

    def __mul__(self, other: Multiplier, /) -> Multiplier:
        """Multiply this multiplier by another"""
        if not self.canmultiplyby(other):
            raise ArithmeticError(f"Can't multiply {self!r} by {other!r}")
        return Multiplier(self.min * other.min, self.max * other.max)

    def __add__(self, other: Multiplier, /) -> Multiplier:
        """Add two multipliers together"""
        return Multiplier(self.min + other.min, self.max + other.max)

    def __sub__(self, other: Multiplier, /) -> Multiplier:
        """
        Subtract another multiplier from this one.
        Caution: multipliers are not totally ordered.
        This operation is not meaningful for all pairs of multipliers.
        """
        mandatory = self.mandatory - other.mandatory
        optional = self.optional - other.optional
        return Multiplier(mandatory, mandatory + optional)

    def canintersect(self, other: Multiplier, /) -> bool:
        """
        Intersection is not well-defined for all pairs of multipliers.
        For example:
            {2,3} & {3,4} = {3}
            {2,} & {1,7} = {2,7}
            {2} & {5} = ERROR
        """
        return not (self.max < other.min or other.max < self.min)

    def __and__(self, other: Multiplier, /) -> Multiplier:
        """
        Find the intersection of two multipliers: that is, a third
        multiplier expressing the range covered by both of the originals.
        This is not defined for all multipliers since they may not overlap.
        """
        if not self.canintersect(other):
            raise ArithmeticError(
                f"Can't compute intersection of {self!r} and {other!r}"
            )
        a = max(self.min, other.min)
        b = min(self.max, other.max)
        return Multiplier(a, b)

    def canunion(self, other: Multiplier, /) -> bool:
        """
        Union is not defined for all pairs of multipliers.
        E.g. {0,1} | {3,4} -> nope
        """
        return not (self.max + Bound(1) < other.min or other.max + Bound(1) < self.min)

    def __or__(self, other: Multiplier, /) -> Multiplier:
        """
        Find the union of two multipliers: that is, a third multiplier
        expressing the range covered by either of the originals. This is
        not defined for all multipliers since they may not intersect.
        """
        if not self.canunion(other):
            raise ArithmeticError(f"Can't compute the union of {self!r} and {other!r}")
        a = min(self.min, other.min)
        b = max(self.max, other.max)
        return Multiplier(a, b)

    def common(self, other: Multiplier, /) -> Multiplier:
        """
        Find the shared part of two multipliers. This is the largest
        multiplier which can be safely subtracted from both the originals.
        This may return the `ZERO` multiplier.
        """
        mandatory = min(self.mandatory, other.mandatory)
        optional = min(self.optional, other.optional)
        return Multiplier(mandatory, mandatory + optional)

    def copy(self, /) -> Multiplier:
        return Multiplier(self.min.copy(), self.max.copy())


# Preset multipliers. These get used ALL THE TIME in unit tests
ZERO = Multiplier(Bound(0), Bound(0))  # has some occasional uses
QM = Multiplier(Bound(0), Bound(1))
ONE = Multiplier(Bound(1), Bound(1))
STAR = Multiplier(Bound(0), INF)
PLUS = Multiplier(Bound(1), INF)

# Symbol lookup table for preset multipliers.
symbolic: Mapping[Multiplier, str] = {
    QM: "?",
    ONE: "",
    STAR: "*",
    PLUS: "+",
}
